program dcci;

{$APPTYPE CONSOLE}

uses
	Windows, SysUtils, StrUtils;

function ExecuteCommandLine(
	const Command: WideString;
	const Directory: WideString;
	const StdIn: string; out StdOut: string; StdErr: PString;
	ErrorLevel: PDWORD): Boolean; 
const
	PipeSecurity: TSecurityAttributes = (
		nLength: SizeOf(TSecurityAttributes);
		lpSecurityDescriptor: nil;
		bInheritHandle: True);
var
	StdInR, StdInW: THandle;
	StdOutR, StdOutW: THandle;
	StdErrR, StdErrW: THandle;
	SInfo: STARTUPINFOW;
	PInfo: PROCESS_INFORMATION;
	Buffer: array[0..1024] of WideChar;
	Written, Read: DWORD;
	From: Integer;
	Terminated: Boolean;
begin
	CreatePipe(StdInR, StdInW, @PipeSecurity, 0);
	CreatePipe(StdOutR, StdOutW, @PipeSecurity, 0);
	if StdErr <> nil then CreatePipe(StdErrR, StdErrW, @PipeSecurity, 0);

	Windows.DuplicateHandle(GetCurrentProcess, StdInW, GetCurrentProcess,
		@StdInW, 0, False, DUPLICATE_SAME_ACCESS or DUPLICATE_CLOSE_SOURCE);
	try
		GetStartupInfoW(STARTUPINFOA(SInfo));
		SInfo.dwFlags := STARTF_USESTDHANDLES or STARTF_FORCEOFFFEEDBACK or STARTF_USESHOWWINDOW;
		SInfo.hStdInput := StdInR;
		SInfo.hStdOutput := StdOutW;
		if StdErr <> nil then SInfo.hStdError  := StdErrW else SInfo.hStdError  := StdOutW;
		SInfo.wShowWindow := SW_MINIMIZE;

		Result := CreateProcessW(nil, PWideChar(Command), @PipeSecurity, nil, True,
			0, nil, PWideChar(Directory), SInfo, PInfo);
		if Result then
		begin
			StdOut := '';
			if StdErr <> nil then StdErr^ := '';
			try
				WaitForInputIdle(PInfo.hProcess, 1000);
				WriteFile(StdInW, Pointer(StdIn)^, Length(StdIn), Written, nil);
				repeat
					Terminated := WaitForSingleObject(PInfo.hProcess, 100) = WAIT_OBJECT_0;

					PeekNamedPipe(StdOutR, nil, 0, nil, @Written, nil);
					while Written > 0 do
					begin
						ReadFile(StdOutR, Buffer, SizeOf(Buffer), Read, nil);
						From := Length(StdOut);
						SetLength(StdOut, From + Integer(Read));
						Move(Buffer, (PChar(StdOut) + From)^, Read);
						Dec(Written, Read);
					end;

					if StdErr <> nil then
					begin
						PeekNamedPipe(StdErrR, nil, 0, nil, @Written, nil);
						while Written > 0 do
						begin
							ReadFile(StdErrR, Buffer, SizeOf(Buffer), Read, nil);
							From := Length(StdErr^);
							SetLength(StdErr^, From + Integer(Read));
							Move(Buffer, (PChar(StdErr^) + From)^, Read);
							Dec(Written, Read);
						end;
					end
				until Terminated;
			finally
				if ErrorLevel <> nil then
					GetExitCodeProcess(PInfo.hProcess, ErrorLevel^);
				CloseHandle(PInfo.hProcess);
				CloseHandle(PInfo.hThread);
			end;
		end
	finally
		CloseHandle(StdInW);
		CloseHandle(StdInR);
		CloseHandle(StdOutW);
		CloseHandle(StdOutR);
		if StdErr <> nil then
		begin
			CloseHandle(StdErrW);
			CloseHandle(StdErrR);
		end
	end;
end;

var
	TempDir: string;
	TempName: string;
	Index: Cardinal;
	Usings: array of string;
	Requires: array of string;
	Modules: array of HMODULE;

procedure Run(const Script: string);
var
	BaseName, UnitName, PackageName, PackageSource: string;
	ScriptType: (stCode, stConst, stVar, stType, stUses);
	F: Text;
	I, P: Integer;
	CompilerMessage, ErrorMessage: string;
	ErrLevel: DWORD;
begin
	if StartsText('const ', Script) then
		ScriptType := stConst
	else if StartsText('var ', Script) then
		ScriptType := stVar
	else if StartsText('type ', Script) then
		ScriptType := stType
	else if StartsText('uses ', Script) then
		ScriptType := stUses
	else
		ScriptType := stCode;

	BaseName := TempName + Format('%.8x', [Index]);
	UnitName := BaseName + '.pas';
	PackageName := BaseName + 'p';
	PackageSource := PackageName + '.dpk';
	Assign(F, TempDir + UnitName);
	Rewrite(F);
	WriteLn(F, 'unit ', BaseName, ';');
	WriteLn(F, 'interface');
	if ScriptType <> stUses then
	begin
		WriteLn(F, 'uses');
		Write(F, 'SysUtils');
		for I := Low(Usings) to High(Usings) do
			Write(F, ',', Usings[I]);
		WriteLn(F, ';');
	end;
	if ScriptType <> stCode then WriteLn(F, Script);
	WriteLn(F, 'implementation');
	if ScriptType = stCode then
	begin
		WriteLn(F, 'begin');
		WriteLn(F, Script);
	end;
	WriteLn(F, 'end.');
	Close(F);
	Assign(F, TempDir + PackageSource);
	Rewrite(F);
	WriteLn(F, 'package ', PackageName, ';');
	WriteLn(F, 'requires');
	Write(F, 'rtl');
	for I := Low(Requires) to High(Requires) do
		Write(F, ',', Requires[I]);
	WriteLn(F, ';');
	WriteLn(F, 'contains');
	WriteLn(F, BaseName, ';');
	WriteLn(F, 'end.');
	Close(F);
	ExecuteCommandLine('dcc32 ' + PackageSource,
		TempDir, '', CompilerMessage, nil, @ErrLevel);
	if ErrLevel = 0 then
	begin
		SetLength(Modules, Length(Modules) + 1);
		Modules[High(Modules)] := LoadPackage(TempDir + PackageName + '.bpl');
		SetLength(Usings, Length(Usings) + 1);
		Usings[High(Usings)] := BaseName;
		SetLength(Requires, Length(Requires) + 1);
		Requires[High(Requires)] := PackageName;
		Inc(Index);
	end
	else
	begin
		CompilerMessage := ReplaceStr(CompilerMessage, #13, #10);
		while CompilerMessage <> '' do
		begin
			if StartsText(UnitName, CompilerMessage) then
			begin
				Delete(CompilerMessage, 1, Length(UnitName));
				I := Pos(#10, CompilerMessage);
				ErrorMessage := Trim(Copy(CompilerMessage, 1, I));
				if ErrorMessage <> '' then
				begin
					if ErrorMessage[1] = '(' then
					begin
						P := Pos(')', ErrorMessage);
						Delete(ErrorMessage, 1, P);
					end;
					if ErrorMessage <> '' then WriteLn(ErrorMessage);
				end;
				Delete(CompilerMessage, 1, I);
			end
			else
			begin
				I := Pos(#10, CompilerMessage);
				Delete(CompilerMessage, 1, I);
			end;
		end;
	end;
end;

var
	TempBuf: array[0..1023] of Char;
	TempLen: Cardinal;
	Param, Script: string;
	FD: TSearchRec;
	I: Integer;
	Mode: (mOneLiner, mIntaractive);
begin
	TempLen := GetTempPathA(Length(TempBuf), TempBuf);
	SetString(TempDir, TempBuf, TempLen);
	TempName := Format('dcci%.8x', [GetCurrentProcessId]);
	Index := 1;

	Mode := mIntaractive;
	for I := 1 to ParamCount do
	begin
		Param := ParamStr(I);
		if Param <> '' then
		begin
			Script := Param;
			Mode := mOneLiner;
		end;
	end;

	case Mode of
	mIntaractive:
		repeat
			repeat
				Write('> ');
				ReadLn(Script);
			until Trim(Script) <> '';
			if SameText(Script, 'exit') then Break;
			Run(Script);
		until False;
	mOneLiner:
		Run(Script);
	end;

	for I := High(Modules) downto Low(Modules) do
		UnloadPackage(Modules[I]);
	if FindFirst(TempDir + TempName + '*', faAnyFile, FD) = 0 then
	begin
		repeat
			DeleteFile(TempDir + FD.Name);
		until FindNext(FD) <> 0;
		FindClose(FD);
	end;
end.
